#!/usr/bin/env python3
import sys 
import argparse

from mashpit import __version__
from mashpit import build 
from mashpit import update
from mashpit import query
#from mashpit import webserver


def commandToArgs(commandline):
    parser = argparse.ArgumentParser(description="A sketch-based surveillance platform.")
    parser.add_argument('-v', '--version',
                       action='version',
                       version='%(prog)s ' + __version__
                       )
    
    # subparsers for mashpit
    subparsers = parser.add_subparsers(metavar='command')
    subparsers.required = True
    subparser_build = subparsers.add_parser("build", help='Build mashpit database')
    subparser_update = subparsers.add_parser("update", help='Update a database')
    subparser_query = subparsers.add_parser("query", help='Query for the most similar isolates')
    subparser_webserver = subparsers.add_parser("webserver", help='Start a local web server')

    # arguments for mashpit build
    subparser_build.add_argument("type", type=str, help="mashpit database type.",choices=['taxon', 'accession'])
    subparser_build.add_argument("name", type=str, help="mashpit database name")
    subparser_build.add_argument("--quiet", action="store_true", help="disable logs")
    subparser_build.add_argument("--number", type=int,default=1000,help="maximum number of hashes for sourmash, default is 1000")
    subparser_build.add_argument("--ksize", type=int,default=31,help="kmer size for sourmash, default is 31")
    subparser_build.add_argument("--species", type=str, help="species name")
    subparser_build.add_argument("--email",type=str,help='Entrez email')
    subparser_build.add_argument("--key",type=str,help='Entrez api key')
    subparser_build.add_argument("--pd_version",type=str,help='a specified Pathogen Detection version (PDG accession). Default is the latest.')
    subparser_build.add_argument("--list",type=str,help='Path to a list of NCBI BioSample accessions')
    subparser_build.set_defaults(func=build.build)

    # "update" subcommand
    # TODO: remove the name argument? because the name is already in the database folder
    subparser_update.add_argument(dest="database", type=str, help="path for the database folder")
    subparser_update.add_argument(dest="name", type=str, help="database name")
    subparser_update.add_argument("--metadata", type=str, help="metadata file in csv format")
    subparser_update.add_argument("--quiet", action="store_true", help="disable logs")
    subparser_update.set_defaults(func=update.update)
    
    # "query" subcommand
    subparser_query.add_argument(dest="sample", type=str, help="file path to the query sample")
    subparser_query.add_argument(dest="database", type=str, help="path to the database folder")
    subparser_query.add_argument("--number", type=int, help="number of isolates in the query output, default is 200",default=200)
    subparser_query.add_argument("--threshold", type=float, help="minimum jaccard similarity for mashtree, default is 0.85",default=0.85)
    subparser_query.add_argument("--annotation", type=str, help="mashtree tip annoatation, default is none")
    subparser_query.set_defaults(func=query.query)

    # "webserver" subcommand
    # subparser_webserver.set_defaults(func=webserver.webserver)

    args = parser.parse_args(commandline)
    return args

def main():
    if len(sys.argv[1:])<1:
        print("Subcommand is required to run mashpit. Use -h or --help to show help information.\n")
        print("Subcommand options:\n")
       
        exit(0)
    args = commandToArgs(sys.argv[1:])
    args.func(args)
    return 



#!/usr/bin/env python3

from setuptools import setup

with open("README.md", "r") as rm:
    long_description = rm.read()

setup(
    name='mashpit',
    version='0.9.4',
    url='https://github.com/tongzhouxu/mashpit',
    author='Tongzhou Xu',
    author_email='tongzhou.xu@uga.edu',
    description='A sketch-based surveillance platform',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['mashpit'],
    include_package_data=True,
    entry_points={'console_scripts':['mashpit=mashpit.mashpit:main']},
    install_requires=[
        'sourmash~=4.6.1',
        'ncbi-datasets-pylib~=14.6.2',
        'pandas',
        'biopython',
        'scikit-bio',
        'python-dotenv',
        'tqdm',
        'flask',
        'dask[dataframe]',
    ]
)



import os
import time
import pandas as pd
import glob
import urllib.request
import logging
import shutil

from sourmash import load_file_as_signatures,save_signatures,load_one_signature
from mashpit.build import create_connection,PDGParser,download_metadata,calculate_centroid,import_metadata,download_and_sketch_assembly

def check_database_version(db_pathogen_name,db_version):
    # check if there is a new version
    url_metadata_folder = 'https://ftp.ncbi.nlm.nih.gov/pathogen/Results/'+db_pathogen_name+'/latest_snps/Metadata/'
    response = urllib.request.urlopen(url_metadata_folder)
    metadata_html = response.read().decode('utf-8')
    parser = PDGParser()
    parser.feed(metadata_html)
    metadata_file_name = parser.list[0]
    latest_pdg_acc = metadata_file_name.strip('.metadata.tsv')
    if db_version == latest_pdg_acc:
        logging.error('%s is already the latest version' % db_version)
        exit(0)
    else:
        logging.info('Updating %s from %s to %s' % (db_pathogen_name,db_version,latest_pdg_acc))
    
    return latest_pdg_acc
    
def compare_tables(database_cluster_center,latest_cluster_center_metadata):
    logging.info('Comparing two databases')
    pds_to_remove = []
    asm_acc_remove = []
    pds_to_add = []
    asm_acc_add = []
    #1. If PDS is no longer in latest metadata, remove old metadata record and signature
    pds_only_in_old = list(set(database_cluster_center['PDS_acc_no_suffix'].to_list()) - set(latest_cluster_center_metadata['PDS_acc_no_suffix'].to_list()))
    if len(pds_only_in_old)>0:
        for pds in pds_only_in_old:
            pds_to_remove.append(pds)
        for pds in database_cluster_center[database_cluster_center['PDS_acc_no_suffix'].isin(pds_to_remove)]['asm_acc'].to_list():
            asm_acc_remove.append(pds)
    #2. If PDS is present in both tables, check if the centroid that has changed
    pds_common = list(set(database_cluster_center['PDS_acc_no_suffix'].to_list()).intersection(latest_cluster_center_metadata['PDS_acc_no_suffix'].to_list()))
    if len(pds_common)>0:
        pds_changed = []
        for pds in pds_common:
            database_center = database_cluster_center[database_cluster_center['PDS_acc_no_suffix']==pds]['PDS_acc'].iloc[0]
            latest_center = latest_cluster_center_metadata[latest_cluster_center_metadata['PDS_acc_no_suffix']==pds]['PDS_acc'].iloc[0]
            if database_center != latest_center:
                pds_changed.append(pds)
        if len(pds_changed)>0:
            for pds in pds_changed:
                pds_to_add.append(pds)
                pds_to_remove.append(pds)
            for pds in database_cluster_center[database_cluster_center['PDS_acc_no_suffix'].isin(pds_changed)]['asm_acc'].to_list():
                asm_acc_remove.append(pds)
                asm_acc_add.append(pds)
    #3. If PDS is only in the new table, add metadata and download signature files
    pds_new = list(set(latest_cluster_center_metadata['PDS_acc_no_suffix'].to_list()) - set(database_cluster_center['PDS_acc_no_suffix'].to_list()))
    if len(pds_new)>0:
        for pds in pds_new:
            pds_to_add.append(pds)
        for pds in latest_cluster_center_metadata[latest_cluster_center_metadata['PDS_acc_no_suffix'].isin(pds_new)]['asm_acc'].to_list():
            asm_acc_add.append(pds)
    
    return pds_to_remove, asm_acc_remove, pds_to_add, asm_acc_add

def delete_metadata(conn, asm_acc):
    sql = 'DELETE FROM METADATA WHERE asm_acc=?'
    cur = conn.cursor()
    try:
        cur.execute(sql, (asm_acc,))
    except:
        logging.error('Unable to delete %s' % asm_acc)
    cur.execute(sql, (asm_acc,))
    conn.commit()

def update_taxon(args,conn,tmp_folder,database_sig_path):
    c = conn.cursor()
    c.execute("SELECT value FROM DESC where name = 'Species';")
    db_pathogen_name = c.fetchone()[0]
    logging.info('Pathogen name: %s' % db_pathogen_name)
    c.execute("SELECT value FROM DESC where name = 'Version';")
    db_version = c.fetchone()[0]
    logging.info('Pathogen Detection version: %s' % db_version)
    c.execute("SELECT value FROM DESC where name = 'Hash_number';")
    hash_number = int(c.fetchone()[0])
    logging.info('Hash number: %s' % hash_number)
    c.execute("SELECT value FROM DESC where name = 'Kmer_size';")
    kmer_size = int(c.fetchone()[0])
    logging.info('Kmer size: %s' % kmer_size)

    latest_pdg_acc = check_database_version(db_pathogen_name,db_version)

    # get the cluster-centroid table from the original database
    database_cluster_center = pd.read_sql_query("select * from METADATA ", conn)
    # calculate the new cluster-centroid table for the new database
    metadata_file_name, isolate_pds_file = download_metadata(db_pathogen_name,latest_pdg_acc, tmp_folder)
    latest_metadata = pd.read_csv(os.path.join(tmp_folder,metadata_file_name),header=0,sep='\t')
    df_isolate_pds = pd.read_csv(os.path.join(tmp_folder,isolate_pds_file),header=0,sep='\t')
    # add cluster accession to metadata
    latest_metadata = latest_metadata.join(df_isolate_pds[['target_acc','PDS_acc']].set_index('target_acc'),on='target_acc')
    latest_metadata_asm = latest_metadata[~latest_metadata['asm_acc'].isnull()]
    calculate_centroid(latest_metadata_asm,latest_pdg_acc,tmp_folder)
    latest_cluster_center = pd.read_csv(os.path.join(tmp_folder,f'{latest_pdg_acc}_cluster_center.tsv'),sep='\t')
    latest_cluster_center_metadata = latest_cluster_center.join(latest_metadata_asm.drop('PDS_acc', axis=1).set_index('target_acc'),on='target_acc')
    # add a new PDS accession column without suffix/PDS version
    database_cluster_center['PDS_acc_no_suffix'] = database_cluster_center['PDS_acc'].str.split('.').str[0]
    latest_cluster_center_metadata['PDS_acc_no_suffix'] = latest_cluster_center_metadata['PDS_acc'].str.split('.').str[0]
    pds_to_remove, asm_acc_remove, pds_to_add, asm_acc_add = compare_tables(database_cluster_center,latest_cluster_center_metadata)
    logging.info('Found %s PDS to remove' % len(pds_to_remove))
    logging.info('Found %s PDS to add' % len(pds_to_add))

    if len(pds_to_remove) > 0:
        for asm_acc in asm_acc_remove:
            delete_metadata(conn, asm_acc)
    if len(pds_to_add) > 0:
        import_metadata(latest_metadata,latest_cluster_center_metadata[latest_cluster_center_metadata['PDS_acc_no_suffix'].isin(pds_to_add)],conn)
        download_and_sketch_assembly(asm_acc_add,hash_number,kmer_size,tmp_folder)
    
    database_sig=load_file_as_signatures(database_sig_path)
    siglist = []
    for sig in database_sig:
        if str(sig) not in asm_acc_remove:
            siglist.append(sig)
    new_sig_path = os.path.join(tmp_folder,'signature',"*.sig")
    sig_path_list = glob.glob(new_sig_path)
    for sig in sig_path_list:
        siglist.append(load_one_signature(sig))
    with open(database_sig_path,'w') as f:
        save_signatures(siglist,fp=f)
    shutil.rmtree(tmp_folder)

    # update the PDS version info in DESC table in the db
    c.execute(f"Update DESC set value = '{latest_pdg_acc}' where name = 'Version';")

    conn.commit()
    conn.close()
    logging.info(f'Database {args.name} updated successfully to version {latest_pdg_acc}')

def update_accession(args,conn,tmp_folder):
    # update the metadata or add new columns
    metadata_file = args.metadata
    if not os.path.exists(metadata_file):
        logging.error('Metadata file not found.')
        exit(1)
    # check if the metadata file has the biosample_acc columns
    metadata = pd.read_csv(metadata_file,header=0)
    if 'biosample_acc' not in metadata.columns:
        logging.error('Metadata file does not have biosample_acc column.')
        exit(1)
    # check if other columns are present in the database
    c = conn.cursor()
    c.execute("SELECT * FROM METADATA")
    columns = [description[0] for description in c.description]
    for col in metadata.columns:
        if col not in columns:
            # add the column to the database
            c.execute(f"ALTER TABLE METADATA ADD COLUMN '{col}'")
    conn.commit()
    # import the metadata
    import_metadata(metadata,conn)

def update(args):
    t = time.localtime()
    current_time = time.strftime("%Y%m%d%H%M%S", t)
    print_handler = logging.StreamHandler()
    file_handler = logging.FileHandler("mashpit-"+current_time+".log")
    print_handler.setLevel(logging.INFO)
    file_handler.setLevel(logging.DEBUG)
    if args.quiet:
        print_handler.setLevel(logging.ERROR)
    logging.basicConfig(format='%(asctime)s [%(levelname)s] %(message)s', 
                        level=logging.DEBUG,
                        datefmt='%d-%b-%y %H:%M:%S', 
                        handlers=[print_handler,file_handler])
    
    db_folder = os.path.abspath(args.database)
    # create a tmp folder
    cwd = os.getcwd()
    tmp_folder = os.path.join(cwd, 'tmp')
    if os.path.isdir(tmp_folder):
        shutil.rmtree(tmp_folder)
    os.mkdir(tmp_folder)
    if not os.path.exists(db_folder):
        logging.error('Path not found.')
        exit(1)
    sql_path = os.path.join(db_folder,args.name + '.db')
    sig_path = os.path.join(db_folder, args.name + '.sig')
    if not (os.path.exists(sql_path)):
        logging.error(f'{sql_path} not found.')
        exit(1)
    if not (os.path.exists(sig_path)):
        logging.error(f'{sig_path} not found.')
        exit(1)

    conn = create_connection(sql_path)
    logging.info('Updating database %s' % args.name)
    c = conn.cursor()
    c.execute("SELECT value FROM DESC where name = 'Type';")
    db_type = c.fetchone()[0]
    logging.info('Database type: %s' % db_type)
    if db_type == 'Taxonomy':
        update_taxon(args,conn,tmp_folder,sig_path)
    elif db_type == 'Accession':
        update_accession(args,conn,tmp_folder)


from flask import Flask, render_template, request,redirect, url_for
from mashpit.query import get_query_sig, generate_query_table, generate_mashtree
import os
import threading


app = Flask(__name__, template_folder=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates'))

# Global variable to store the output of the classification
mashpit_out = None
mashpit_thread = None
mashpit_status = "in progress"

def run_mashpit(database_db,database_sig,assembly_file):
    return


@app.route("/", methods=['GET'])
def upload():
    return render_template('upload.html')

@app.route("/submit", methods=['POST'])
def submit():
    database_db = request.files['formFile_db']
    file_path_db = os.path.join(app.config['UPLOAD_FOLDER'], database_db.filename)
    database_db.save(file_path_db)
    database_sig = request.files['formFile_sig']
    file_path_sig = os.path.join(app.config['UPLOAD_FOLDER'], database_sig.filename)
    database_sig.save(file_path_sig)
    assembly_file = request.files['formFile_assembly']
    file_path = os.path.join(app.config['UPLOAD_FOLDER'], assembly_file.filename)
    assembly_file.save(file_path)

    global mashpit_thread
    global mashpit_status
    
    mashpit_thread = threading.Thread(target=run_mashpit, args=(database_db,database_sig,assembly_file,))
    mashpit_thread.start()

    return redirect(url_for('loading'))

@app.route('/loading')
def loading():
    """
    Show loading page while classification is running
    """
    global mashpit_thread
    global mashpit_status
    return render_template('loading.html')

@app.route('/result')
def result():
    """
    Show classification results
    """
    global mashpit_out
    global mashpit_status
    mashpit_status = "in progress"
    return render_template('result.html', table=mashpit_out)

@app.route('/status')
def status():
    """
    Return the current status of the classification
    """
    global mashpit_status
    return mashpit_status

def webserver(args):
    app.config['UPLOAD_FOLDER'] = 'uploads'
    os.makedirs(app.config['UPLOAD_FOLDER'], exist_ok=True)
    app.run(debug=True)
    